from jax.config import config

config.update("jax_enable_x64", False)

import matplotlib
import numpy as np
from scipy import stats

import opt_algorithms

matplotlib.use('Agg')
import matplotlib.pyplot as plt
import pickle
import importlib
import matplotlib.cm as cm
import jax.scipy.stats.norm as jaxnm
from matplotlib.patches import Circle
import jax.numpy as jnp
from datetime import datetime
from scipy.spatial import ConvexHull
from jax import vmap,jit
import dimmer_utils
import func_utils
import potentials_2d
import time
import sys
import math
import dimmer
import jax_md
import hungarian
# import rmsd

importlib.reload(potentials_2d)
importlib.reload(func_utils)
importlib.reload(dimmer_utils)
importlib.reload(dimmer)

def calcAdjMatrices(x,y,N,beta):
  # beta=1
  Adj_x=np.zeros((N,N))
  Adj_y=np.zeros((N,N))

  for k in range(N):
  # for l in range(prms.N_a):
    Adj_x[k]=np.sqrt(np.sum((x[k,:]-x)*(x[k,:]-x),axis=1))**2
    Adj_x[k]=np.exp(Adj_x[k]/beta)
    # np.linalg.norm(di[k]-di[l])
    Adj_y[k]=np.sqrt(np.sum((y[k,:]-y)*(y[k,:]-y),axis=1))**2
    Adj_y[k]=np.exp(Adj_y[k]/beta)
  return Adj_x,Adj_y

def calc_distance_hungarian(x,y,N_a,beta):
    Adj_x,Adj_y=calcAdjMatrices(x, y,N_a,beta)
    mu_x,v_x=np.linalg.eigh(Adj_x)
    mu_y,v_y=np.linalg.eigh(Adj_y)
    M=np.abs(v_x)@np.abs(v_y.T)
    costMatrix = hungarian.Hungarian.make_cost_matrix(M)
    h_cost = hungarian.Hungarian(costMatrix)
    h_cost.calculate()
    opt_dist=h_cost.get_total_potential()
    return opt_dist

def plot_2dLJ(X_t, prms, axis, cols=None, t=0, fEval=0, grad_norm=0, first_egv=0, first_pos=0):
    X_t=X_t-X_t[0]
    for p in range(0, prms.N_a):
        if cols is None:
            at = Circle((X_t[:, 0][p], X_t[:, 1][p]), math.pi / 6, alpha=0.75, color='grey')
        else:
            at = Circle((X_t[:, 0][p], X_t[:, 1][p]), math.pi / 6, alpha=0.75, color=cols[p])
            axis.annotate(str(p), xy=(X_t[:, 0][p], X_t[:, 1][p]), fontsize=20, ha="center", va="center")
        axis.add_patch(at)
        # axis.text(X_t[:, 0][p], X_t[:, 1][p], str(p), fontsize=20, fontweight='bold', color='k')

    title = 't={:<3.1f},V={:<12.5f},dV={:<1.5f},l0={:<1.5f},l1={:<1.5f}'.format(t, fEval,grad_norm,first_egv,first_pos)

    axis.title.set_text(title)
    axis.autoscale_view()
    return axis


def plot_text(ax_t,prms,xs,ys):
    plot_params=make_parameter_string(prms)
    ax_t.text(xs,ys,plot_params,fontsize=9)
    plt.tight_layout()
    plt.savefig(prms.str_file+str(0)+str(".png"),dpi=100)

def get_stats(X,dv,ddV,total_d):
    dv_norm = jnp.linalg.norm(dv(X))
    hess_eval = ddV(X).reshape(total_d, total_d)
    mu, v = np.linalg.eigh(hess_eval)
    return dv_norm,hess_eval,mu,v

def make_parameter_string(prms):

  now = datetime.now()
  dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
  plot_str = "time:"+dt_string+"\n"
  dicprms=vars(prms)
  listprms=list(dicprms)
  for i in listprms:
#      key_name=listprms[i]
      plot_str+=str(i)+'='
      plot_str+=str(dicprms[i])+'\n'
  return plot_str

# load parameters
def vis_lj():
    plt.rcParams.update({
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": ["Helvetica"]})
    # for Palatino and other serif fonts use:
    plt.rcParams.update({
        "text.usetex": True,
        "font.family": "serif",
        "font.serif": ["Palatino"],
    })
    plt.figure(figsize=(1200, 1200))
    plt.rcParams.update({'font.size': 3})
    #
    plt.rc('xtick', labelsize=6)
    plt.rc('ytick', labelsize=6)
    fig, axs = plt.subplots(1,1)
    axs.set_xlim(-2, 2)
    axs.set_ylim(-2, 2)
    axs.set_aspect('equal', 'box')

    file_name="../lj7_1/lj7_prms.pickle"
    file = open(file_name, 'rb')
    plot_index1=True
    prms=pickle.load(file)
    file.close()
    prms,V,V_plot,x_s, y_s, l_s,X0,limits=potentials_2d.setup_problem(prms)
    total_d=prms.d*prms.N_a
    plot_text([axs][0],prms,-2,-2)
    plt.savefig(prms.str_file+str("_prms.png"), dpi=100)
    plt.rcParams.update({'font.size': 6})
    plt.figure(figsize=(1200, 1200))
    fig, axs = plt.subplots(2,3)
    for i in range(0,2):
        for j in range(0, 3):
            axs[i,j].set(xlim=(-2, 2), ylim=(-2, 2))
            axs[i,j].set_aspect('equal', 'box')
    fig.tight_layout()




    T=np.int32(prms.T_real/prms.h_x)
    colors = cm.gist_rainbow(np.linspace(0, 1, prms.N_a))
    Vm,dV_s,ddV_s,dV,hvp_vec,ddV,index_func_lambda,index_func_lambda_map,gd_map=func_utils.get_func_evals(V,prms)
    dv_norm_0,hess_eval,mu_0,v_0=get_stats(X0,dV_s,ddV_s,total_d)

    # axs[0]=plot_2dLJ(X0, prms, axs[0], colors, t=0, fEval=V(X0), grad_norm=dv_norm, first_egv=mu[0], first_pos=mu[1+prms.eig_gap])
    # plt.savefig(prms.str_file + str("0.png"), dpi=100)


    prms.gd_dt=0.001
    prms.gd_max_iter=5000
    prms.basin_gd_tol=10e-8

    list_linked_saddles=None
    list_other_saddles=None
    list_minima=jnp.array(V(X0))


    # plot inital x configuration
    # N_f=1000
    N_f=1000
    for fs in range(1,N_f):
        # if fs%100==0:
        #     print(fs, "/", N_f)

        t1 = time.perf_counter()
        # load states
        # print(fs)
        file_x = open(prms.str_file + "x_stats_" + str(fs) + ".pickle", "rb")
        [out_s,t,X_t,Y_t,wgts,run_dimmer,S_card]=pickle.load(file_x)

        file_x.close()
        for i in range(0, 2):
            for j in range(0, 3):
                axs[i, j].clear()
                axs[i, j].set(xlim=(-2.5, 2.5), ylim=(-2.5, 2.5 ))
                axs[i, j].set_aspect('equal', 'box')
        linked_saddle=False
        other_saddle = False
        if run_dimmer:
            file_d = open(prms.str_file + "d_stats_" + str(fs) + ".pickle", "rb")
            [U_0, t, exit_code, dimmer_iters, x_index1, x_final] = pickle.load(file_d)

            file_d.close()
            # if sucess dimer
            if exit_code==0:
                # print(t, V(X0), V(U_0), V(x_index1))
                #check if initial condition for dimmer U0 goes back to local minimum
                #was the point in the same basin of attraction as the original point?
                success,x_k,vEval,dVNorm=opt_algorithms.gd_vanilla(V,dV_s,prms.gd_dt ,prms.gd_max_iter,prms.basin_gd_tol,U_0)

                # x_k=1+X0
                # x_k=U_0
                diff_x0=jnp.linalg.norm(x_k-X0)
                # opt_dist=rmsd.calculate_rmsd.rmsd(X0,x_k)
                opt_dist=calc_distance_hungarian(x_k, X0, N_a=prms.N_a, beta=1)
                # print(success,dVNorm,vEval,diff_x0,opt_dist)
                # saddle point associated with initial point
                if opt_dist<0.1:
                    linked_saddle = True
                    if list_linked_saddles is None:
                        list_linked_saddles = jnp.array([V(x_index1)])
                    elif jnp.min(jnp.abs(list_linked_saddles-V(x_index1)))>10e-5:
                        list_linked_saddles =jnp.append(list_linked_saddles, V(x_index1))

                else:
                    other_saddle = True
                    if list_other_saddles is None:
                        list_other_saddles = jnp.array([V(x_index1)])
                    elif jnp.min(jnp.abs(list_other_saddles - V(x_index1))) > 10e-5:
                        list_other_saddles =jnp.append(list_other_saddles, V(x_index1))


                # intial configuarion
                axs[0,0] = plot_2dLJ(X0, prms, axs[0,0], colors, t=0, fEval=V(X0), grad_norm=dv_norm_0, first_egv=mu_0[0], first_pos=mu_0[1+prms.eig_gap])
                axs[0,0].set_xlabel('Initial Point-X0')
                # dimmer inital point
                dv_norm_u0, hess_eval, mu_u0, v_u0 = get_stats(U_0, dV_s, ddV_s, total_d)
                axs[0, 1] = plot_2dLJ(U_0, prms, axs[0, 1], colors, t=t, fEval=V(U_0), grad_norm=dv_norm_u0, first_egv=mu_u0[0], first_pos=mu_u0[1 + prms.eig_gap])
                axs[0, 1].set_xlabel('Dimer Initial Point-U0')
                # initial point basin of attraction
                dv_norm_xk, hess_eval, mu_xk, v_xk = get_stats(x_k, dV_s, ddV_s, total_d)
                axs[1, 0] = plot_2dLJ(x_k, prms, axs[1, 0], colors, t=0, fEval=V(x_k), grad_norm=dv_norm_xk, first_egv=mu_xk[0], first_pos=mu_xk[1 + prms.eig_gap])
                str_title = 'GD(U0) - U_m,hg dist.={:<0.3f},l2 dist.={:<0.3f}'.format(opt_dist,diff_x0)
                axs[1, 0].set_xlabel(str_title)

                # saddle point
                dv_norm_x_index1, hess_eval, mu_x_index1, v_x_index1= get_stats(x_index1, dV_s, ddV_s, total_d)
                axs[1, 1] = plot_2dLJ(x_index1, prms, axs[1, 1], colors, t=0, fEval=V(x_index1), grad_norm=dv_norm_x_index1, first_egv=mu_x_index1[0], first_pos=mu_x_index1[1 + prms.eig_gap])
                axs[1, 1].set_xlabel('Dimer Final Point-UT')

                # find minimum after a positive perturbation
                d_nudge =0.15 * (V(X0) - V(x_index1))* (v_x_index1[:, 0]).reshape(prms.N_a, prms.d)
                X_plus = x_index1 +  d_nudge
                X_minus= x_index1 -  d_nudge
                success_p, x_plus, vEval_plus, dVNorm_plus = opt_algorithms.gd_vanilla(V, dV_s, prms.gd_dt, prms.gd_max_iter, prms.basin_gd_tol, X_plus)
                success_m, x_minus, vEval_minus, dVNorm_minus = opt_algorithms.gd_vanilla(V, dV_s, prms.gd_dt, prms.gd_max_iter, prms.basin_gd_tol, X_minus)
                # print(success_p,success_m,vEval_plus,vEval_minus)

                if (success_p + success_m + success) !=3:
                    print("Check GD parameters, something not converging")
                    sys.exit(-1)
                else:
                    # if minima are unique add the to the list
                    if jnp.min(jnp.abs(list_minima-vEval_plus)>prms.basin_gd_tol):
                        list_minima = jnp.append(list_minima,vEval_plus)
                    if jnp.min(jnp.abs(list_minima-vEval_minus)>prms.basin_gd_tol):
                        list_minima = jnp.append(list_minima,vEval_minus)
                    dv_norm_plus, hess_eval, mu_x_plus, v_x_plus = get_stats(x_plus, dV_s, ddV_s, total_d)
                    axs[0, 2] = plot_2dLJ(x_plus, prms, axs[0, 2], colors, t=0, fEval=vEval_plus, grad_norm=dv_norm_plus, first_egv=mu_x_plus[0], first_pos=mu_x_plus[1 + prms.eig_gap])
                    axs[0, 2].set_xlabel('GD(UT+Nudge)')

                    dv_norm_minus, hess_eval, mu_x_minus, v_x_minus = get_stats(x_minus, dV_s, ddV_s, total_d)
                    axs[1, 2] = plot_2dLJ(x_minus, prms, axs[1, 2], colors, t=0, fEval=vEval_minus, grad_norm=dv_norm_minus, first_egv=mu_x_minus[0], first_pos=mu_x_minus[1 + prms.eig_gap])
                    axs[1, 2].set_xlabel('GD(UT-Nudge)')

                if other_saddle:

                    print(t,"saddle no=",len(list_other_saddles)," (other) saddle ",V(x_index1),"Min+=",vEval_plus,"Min-=",vEval_minus,"dist to original (hng)=",opt_dist," (l2) =",diff_x0)

                if linked_saddle:
                    print(t,"saddle no=",len(list_linked_saddles),"(linked) saddle ",V(x_index1),"Min+=",vEval_plus,"Min-=",vEval_minus,"dist to original=",opt_dist," (l2) =",diff_x0)

                if other_saddle or linked_saddle:
                    plt.savefig(prms.str_file + str(fs) + ".svg", dpi=600, format='svg')
                    plt.savefig(prms.str_file + str(fs) + ".png", dpi=600, format='png')


            # prms.dimmer_dt=0.0001
            # exit_code, dimmer_iters, x_index1, x_final = \
            #     dimmer.ideal_dimmer(U_0, V, dV_s, ddV_s, beta=2.0, dt=prms.dimmer_dt, eig_eps=prms.eig_eps, eps=prms.dimmer_eps, max_iter=prms.dimmer_max_iter, eig_gap=prms.eig_gap,debug=True)
            # print(exit_code)

            # dv_norm = jnp.linalg.norm(dV_s(U_0))
            # hess_eval = ddV_s(U_0).reshape(total_d, total_d)
            # mu, v = np.linalg.eigh(hess_eval)
            # axs[0] = plot_2dLJ(U_0, prms, axs[0], colors, t=0, fEval=V(U_0), grad_norm=dv_norm, first_egv=mu[0], first_pos=mu[1 + prms.eig_gap])
            #
            # dv_norm = jnp.linalg.norm(dV_s(x_final))
            # hess_eval = ddV_s(x_final).reshape(total_d, total_d)
            # mu, v = np.linalg.eigh(hess_eval)
            # axs[1] = plot_2dLJ(x_final, prms, axs[1], colors, t=0, fEval=V(x_final), grad_norm=dv_norm, first_egv=mu[0], first_pos=mu[1 + prms.eig_gap])
            #
            # plt.savefig(prms.str_file + str(fs)+".png", dpi=100)


    #



